import click
import re


def read_file(file_path):
    try:
        with open(file_path, 'r', encoding='utf-8') as file:
            file_contents = file.read()
        return file_contents
    except FileNotFoundError:
        click.echo(f"Error: The file '{file_path}' was not found.")
        return None
    except Exception as e:
        click.echo(f"An error occurred while reading the file '{file_path}': {e}")
        return None


def table_extraction_regex(sql_content):

    table_part_regex_str = (
        r"(" 
            # Checks for table_{{ data_interval_start.format("YYYYMMDD") }}
            r"([a-zA-Z0-9_.-]+?)_{{ data_interval_start\.format\([\'\"]YYYYMMDD[\'\"]\) }}"
            # Checks for table_{{ data_interval_end.format("YYYYMMDD") }}
            r"|([a-zA-Z0-9_.-]+?)_{{ data_interval_end\.format\([\'\"]YYYYMMDD[\'\"]\) }}"
            # Checks for generic table name
            r"|({{.*?}})"
            # Checks for literal table name e.g. my_table, sales_2023 (keep at end of checks)
            r"|([a-zA-Z0-9_.-]+)"
        r")"
    )

    table_pattern = re.compile(
        r"(?:FROM|JOIN)\s+\`"
        r"((?:{{.*?}}|[a-zA-Z0-9_.-]+))"  # project_id check
        r"\."
        r"((?:{{.*?}}|[a-zA-Z0-9_.-]+))"  # datase_id check
        r"\."
        + table_part_regex_str +
        r"\`",
        re.IGNORECASE
    )

    return table_pattern.findall(sql_content)


def write_output(file_path, index):
    try:
        with open(file_path, 'w', encoding='utf-8') as file:
            file.write(index)
        click.echo(f"Successfully wrote content to '{file_path}'")
        return True
    except Exception as e:
        click.echo(f"Error: Could not write to file '{file_path}': {e}")
        return False


import requests

from awehflow.alerts.base import Alerter
from airflow.hooks.base import BaseHook

class MSTeamsAlerter(Alerter):
    """Sends an adaptive card message to microsoft teams"""
    def __init__(self, ms_teams_conn_id: str):
        self.ms_teams_conn_id=ms_teams_conn_id
   
    def alert(self, context):
        
        headers = {"content-type": "application/json"}
        requests.post(
            url=BaseHook.get_connection(self.ms_teams_conn_id).password,
            json=self.__build_message(context), 
            headers=headers
        )

    def __build_message(self, context) -> dict:

        alert_level = context['name'].upper()
        entities = []

        if alert_level == 'FAILURE':
            dag_id = context['body']['name']
            task_id = context['body']['error']['task_id']
            log_url = context['body']['error']['log_url']
            message_body=context['body']['error']['message']
            heading_color="attention"
            if 'engineers' in context['body']:
                engineers=[
                    {
                        'name': engineer['name'],
                        'email': engineer['email'],
                    } for engineer in context['body']['engineers']
                ]
                engineers_fact = ", ".join([
                    f"<at>{ engineer['name'] }</at> ({ engineer['email'] })" 
                    for engineer in engineers
                ])
                entities = [
                    {
                        "type": "mention",
                        "text": f"<at>{ engineer['name'] }</at>",
                        "mentioned": {
                            "id": f"{engineer['email']}",
                            "name": f"{engineer['name']}"
                        }
                    } for engineer in engineers
                ]
            else:
                engineers_fact = "UNKNOWN"

            facts = [
                {
                "title": "Engineers",
                "value": engineers_fact
                },
                {
                "title": "DAG",
                "value": dag_id
                },
                {
                "title": "Task",
                "value": task_id
                },
            ]
            error_message_formatted = {
                                "type": "TextBlock",
                                "style": "default",
                                "size": "default",
                                "weight": "default",
                                "color": "default",
                                "wrap": True,
                                "max_lines": 10,
                                "text": f"{message_body}"
                            }
            
            log_viewer = [
                            {
                                "type": "Action.OpenUrl",
                                "title": "View",
                                "url": f"{log_url}",
                                "role": "button"
                            }
                        ]

        elif alert_level == 'SUCCESS':
            heading_color="good"
            facts = [{'title': key, 'value': context['body'][key]} for key, value in context['body'].items() if
                        key not in ['project', 'engineers', 'status', 'end_time']]
        else:
            heading_color="default"
            facts = [{'title': key, 'value': context['body'][key]} for key, value in context['body'].items() if
                        key not in ['project', 'engineers', 'status', 'end_time']]
    
        message = {
            "type":"message",
            "attachments":[
                {
                    "contentType":"application/vnd.microsoft.card.adaptive",
                    "content":{
                        "$schema":"http://adaptivecards.io/schemas/adaptive-card.json",
                        "type":"AdaptiveCard",
                        "version":"1.2",
                        "body":[
                            {
                                "type": "TextBlock",
                                "style": "heading",
                                "size": "medium",
                                "weight": "bolder",
                                "color": heading_color,
                                "text": alert_level
                            },
                            {
                                "type": "FactSet",
                                "spacing": "large",
                                "facts": facts
                            }
                        ]
                    }
                }
            ]
        }
             
        if len(entities) > 0:
            message["attachments"][0]["content"]["msteams"]={"entities": entities}
        if alert_level == 'FAILURE':
            message["attachments"][0]["content"]["body"].append(error_message_formatted)
            message["attachments"][0]["content"]["actions"] = log_viewer

        return message


import click
import re

from awehflow_cli.config_scripts.utils import read_file, table_extraction_regex, write_output


def generate_dependency_checks(sql):

    matches = table_extraction_regex(sql)

    if not matches:
        return ""

    dependencies_string = "dependencies:"
    unique_tables = set()

    for match_tuple in matches:
        project_id            = match_tuple[0]
        dataset_id            = match_tuple[1]
        table_id              = match_tuple[2]
        start_date_template   = match_tuple[3]
        end_date_template     = match_tuple[4]
        generic_template      = match_tuple[5]
        literal_template      = match_tuple[6]

        date_template_table = None

        if start_date_template:
            date_template_table = start_date_template
            param_table_id_for_yaml = f'{date_template_table}_{{{{ data_interval_start.format("YYYYMMDD") }}}}'
            # Retrieving table name for dependency id
            temp_id_suffix = re.sub(r'[^a-zA-Z0-9_]', '_', date_template_table)
            dependency_id_suffix = re.sub(r'_+', '_', temp_id_suffix).strip('_') or "date_templated_table"

        elif end_date_template:
            date_template_table = end_date_template
            param_table_id_for_yaml = f'{date_template_table}_{{{{ data_interval_end.format("YYYYMMDD") }}}}'
            # Retrieving table name for dependency id
            temp_id_suffix = re.sub(r'[^a-zA-Z0-9_]', '_', date_template_table)
            dependency_id_suffix = re.sub(r'_+', '_', temp_id_suffix).strip('_') or "date_templated_table"

        elif generic_template:
            param_table_id_for_yaml = generic_template
            # Retrieving table name for dependency id
            temp_id = generic_template.replace("{{", "").replace("}}", "")
            temp_id = temp_id.replace("params.", "").replace(".params", "")
            temp_id = temp_id.replace("var.value.", "").replace(".var.value", "")
            temp_id = temp_id.replace("data_interval_start.format", "dtstartfmt")
            temp_id = temp_id.replace("data_interval_end.format", "dtendfmt")
            temp_id = temp_id.strip()
            temp_id = re.sub(r'[^a-zA-Z0-9_]+', '_', temp_id)
            temp_id = re.sub(r'_+', '_', temp_id).strip('_')
            dependency_id_suffix = temp_id if temp_id else "generic_templated_table"

        elif literal_template:
            param_table_id_for_yaml = literal_template
            # Retrieving table name for dependency id
            temp_id_suffix = re.sub(r'[^a-zA-Z0-9_]', '_', literal_template)
            dependency_id_suffix = re.sub(r'_+', '_', temp_id_suffix).strip('_') or "literal_table"

        else:
            click.echo(f"Warning: Could not categorize table string: '{table_id}' from tuple: {match_tuple}")
            param_table_id_for_yaml = table_id
            dependency_id_suffix = "unknown_table_type"

        # Ensuring distinct dependency checks are created
        table_identifier_for_uniqueness = f"{project_id}.{dataset_id}.{table_id}"

        if table_identifier_for_uniqueness not in unique_tables:
            unique_tables.add(table_identifier_for_uniqueness)

            dependencies_string += f"""
  - id: 'dependency_{dependency_id_suffix}'
    operator: 'airflow.contrib.sensors.bigquery_sensor.BigQueryTableSensor'
    params:
      project_id: '{project_id}'
      dataset_id: '{dataset_id}'
      table_id: '{param_table_id_for_yaml}'
    poke_interval: 60
    timeout: 10800
    mode: 'reschedule'"""

    return dependencies_string


def generate_partitioning(partition):
      if not partition:
          return ""
      else:
          partitioning = f"""
      time_partitioning:
        type: '{partition["partition_type"]}'
        field: '{partition["partition_field"]}'"""
      
      return partitioning


def generate_cluster_fields(cluster_fields):
    if not cluster_fields:
      return ""

    items = cluster_fields.split(',')
    processed_items = [item.strip() for item in items if item.strip()]
    if len(processed_items) > 4:
        raise ValueError(f"{len(processed_items)} cluster fields provided which exceeds the maximum allowed of 4.")
    if not processed_items:
      return ""

    output_lines = ["cluster_fields:"]
    for item in processed_items:
      output_lines.append(f"        - '{item}'")
    
    return "\n".join(output_lines)


def generate_config(file, usecase, start_date, schedule, source_project, source_dataset, target_project, target_dataset, partition, cluster_fields):

    sql = read_file(file)
    dependencies = generate_dependency_checks(sql)
    partitioning = generate_partitioning(partition)
    clustering = generate_cluster_fields(cluster_fields)

    index = f"""name: '{usecase}'
start_date: '{start_date}'
catchup: false
schedule: '{schedule}'
version: 1
engineers:
  - name: ''
  - email: ''
alert_on:
  - 'failure'
params:
  default:
    source_project: '{source_project}'
    source_dataset: '{source_dataset}'
    source_dataset_latest: '{source_dataset}_latest'
    target_project: '{target_project}'
    target_dataset: '{target_dataset}'
    target_dataset_latest: '{target_dataset}_latest'
  staging:
    target_project: ''
  production:
    target_project: ''

default_dag_args:
  gcp_conn_id: ''
  write_disposition: 'WRITE_TRUNCATE'
  use_legacy_sql: false
  poke_interval: 1800

{dependencies}

tasks:
  - id: 'create_{usecase}'
    operator: 'awehflow.operators.gcp.BigQueryJobOperator'
    params:
      destination_dataset_table: '{'{{ params.target_project }}.{{ params.target_dataset }}'}.{usecase}_{'{{ data_interval_start.format("YYYYMMDD") }}'}'
      sql: 'configs/{usecase}/{usecase}.sql'
      {partitioning}
      {clustering}

  - id: 'get_latest_{usecase}'
    operator: 'awehflow.operators.gcp.BigQueryPushFirstResultToXComOperator'
    params:
      use_legacy_sql: false
      sql: SELECT MAX(suffix)
        FROM (
          SELECT SUBSTR(table_name, LENGTH('{usecase}_') + 1) as suffix
          FROM `{'{{ params.target_project }}.{{ params.target_dataset }}'}.INFORMATION_SCHEMA.TABLES`
          WHERE table_name LIKE '{usecase}_2%'
        )
    upstream: 
      - 'create_{usecase}'

  - id: 'create_{usecase}_latest'
    operator: 'awehflow.operators.gcp.BigQueryJobOperator'
    params:
      destination_dataset_table: '{'{{ params.target_project }}.{{ params.target_dataset_latest }}'}.{usecase}'
      sql: SELECT * FROM `{'{{ params.target_project }}.{{ params.target_dataset }}'}.{usecase}_'{{{{ ti.xcom_pull(task_ids="get_latest_{usecase}")[0] }}}}'`
    upstream: 
      - 'get_latest_{usecase}'

  - id: 'update_table_descriptions'
    operator: 'awehflow.operators.gcp.BigQueryUpdateTableDescriptionOperator'
    params:
      description_data: 'prepared/configs/{usecase}/field_descriptions.conf'
      destination_dataset_tables: 
        - '{'{{ params.target_project }}.{{ params.target_dataset_latest }}'}.{usecase}'
        - '{'{{ params.target_project }}.{{ params.target_dataset }}'}.{usecase}_{'{{ data_interval_end.format("YYYYMMDD") }}'}'
    upstream: 
      - 'create_{usecase}_latest'

  - id: 'save_bigquery_job_metrics'
    operator: 'awehflow.operators.gcp.BigQueryJobTaskMetricOperator'
    params:
      task_ids:
        - 'create_{usecase}'
        - 'create_{usecase}_latest'
    upstream:
      - 'update_table_descriptions'
"""

    write_output('index.yml', index)


import click
import os
import re
import yaml

from awehflow_cli.config_scripts.utils import read_file, table_extraction_regex


def extract_tables(sql: str) -> set:

    tables = []
    matches = table_extraction_regex(sql)
    for match_tuple in matches:
        project_id            = match_tuple[0]
        dataset_id            = match_tuple[1]
        table_id              = match_tuple[2]
        start_date_template   = match_tuple[3]
        end_date_template     = match_tuple[4]
        generic_template      = match_tuple[5]
        literal_template      = match_tuple[6]

        current_prefix = None
        is_literal = False
        is_generic = False

        if start_date_template:
            current_prefix = start_date_template
        elif end_date_template:
            current_prefix = end_date_template
        elif literal_template:
            is_literal = True
        elif generic_template:
            is_generic = True
            
        tables.append({
            "pdt": f"{project_id}.{dataset_id}.{table_id}",
            "name": table_id,
            "prefix": current_prefix,
            "is_literal": is_literal,
            "is_generic": is_generic
        })
    return tables


def extract_dependencies(index: str) -> set:
    dependencies = {
        "bq_sensor": set(),
        "sql_sensor": set()
    }
    try:
        data = yaml.safe_load(index)
        if not isinstance(data, dict):
            click.echo("Warning: Root of index content is not a dictionary.")
            return dependencies

        dependencies_list = data.get("dependencies", [])
        if not isinstance(dependencies_list, list):
            click.echo("Warning: 'dependencies' key is not a list or is missing.")
            return dependencies

        for dep_item in dependencies_list:
            if isinstance(dep_item, dict):
                operator = dep_item.get("operator", "")
                params = dep_item.get("params", {})

                if "BigQueryTableSensor" in operator:
                    project_id = params.get("project_id")
                    dataset_id = params.get("dataset_id")
                    table_id = params.get("table_id")
                    
                    if project_id and dataset_id and table_id:
                        pdt = f"{project_id}.{dataset_id}.{table_id}"
                        dependencies["bq_sensor"].add(pdt)
                    else:
                        click.echo(f"Warning: Incomplete dependency params in item: {dep_item}")
                elif "SqlSensor" in operator:
                    sql_query = params.get("sql")
                    if isinstance(sql_query, str):
                        # Regex to find "source = 'table_name'" or "source='table_name'"
                        source_matches = re.findall(
                            r"source\s*=\s*\'([a-zA-Z0-9_.-]+)\'",
                            sql_query,
                            re.IGNORECASE
                        )
                        for table_name in source_matches:
                            dependencies["sql_sensor"].add(table_name)

    except yaml.YAMLError as e:
        click.echo(f"Error parsing YAML string: {e}")
    except Exception as e:
        click.echo(f"An unexpected error occurred during YAML processing: {e}")
    return dependencies


def validate_config(configs):
    try:
        config_paths = {}
        for root, dirs, files in os.walk(configs):
            if root == configs:
                for subdir_name in dirs:
                    usecase_path = os.path.join(root, subdir_name)
                    yml_file = None
                    sql_file = None

                    for item in os.listdir(usecase_path):
                        item_path = os.path.join(usecase_path, item)
                        if os.path.isfile(item_path):
                            if item.endswith('.yml'):
                                yml_file = item_path
                            elif item.endswith('.sql'):
                                sql_file = item_path

                    if yml_file and sql_file:
                        config_paths[subdir_name] = {
                            'yml_path': yml_file,
                            'sql_path': sql_file
                        }
                    elif yml_file or sql_file:
                        click.echo(f"Warning: In '{subdir_name}', either .yml or .sql file is missing. Found index: {bool(yml_file)}, SQL: {bool(sql_file)}")
                dirs[:] = []
            else:
                dirs[:] = []

    except Exception as e:
        click.echo(f"Error: Directory '{configs}' not found.")

    missing_dependencies_for_sql_tables = []

    for usecase, file_paths in config_paths.items():

        sql = read_file(file_paths['sql_path'])
        index = read_file(file_paths['yml_path'])

        tables = extract_tables(sql)
        dependencies = extract_dependencies(index)

        if not tables:
            click.echo("INFO: No tables were extracted from the SQL query. Nothing to validate.")
        else:
            for sql_table_info in tables:
                sql_pdt = sql_table_info["pdt"]
                matching = False
                
                if sql_pdt in dependencies["bq_sensor"]:
                    matching = True
                else:
                    base_name_to_check = None
                    if sql_table_info["prefix"]:
                        base_name_to_check = sql_table_info["prefix"]
                    elif sql_table_info["is_literal"]:
                        base_name_to_check = sql_table_info["name"]

                    if base_name_to_check in dependencies["sql_sensor"]:
                        matching = True
                if not matching:
                    missing_dependencies_for_sql_tables.append(sql_pdt)
        
    if matching:
        click.echo("✅ SUCCESS: All tables in the SQL query have corresponding dependency checks.")
    else:
        click.echo("❌ FAILURE: Some tables in the SQL query are missing dependency checks.")
        if missing_dependencies_for_sql_tables:
            click.echo("  The following tables lack dependency checks:")
            for missing_dep in missing_dependencies_for_sql_tables:
                click.echo(f"    - {missing_dep}")



import datetime
import json
import os
from unittest import TestCase
import pendulum

from airflow.models.dag import DAG
from airflow.models.baseoperator import BaseOperator
from airflow.operators.bash import BashOperator
from airflow.operators.python import PythonOperator
from airflow.serialization.serialized_objects import DagDependency, SerializedDAG
from airflow.utils.state import DagRunState
from airflow.utils.types import DagRunType
from airflow.models import TaskInstance

from awehflow.config import Config
from awehflow.core import ConfigCache
from awehflow.core import DagLoader
from awehflow.core import DependencySensors
from awehflow.core import TasksSubDag
from tests.helpers import prepare_clean_tmp_dir
from tests.helpers import tmp_dir_path
from tests.helpers import ensure_dir_exists
from tests.helpers import dict_to_yml
from tests.helpers import dict_to_hocon


class TestConfigCache(TestCase):

    def setUp(self):
        self.yaml_configs = [
            {
                'name': 'pipeline_1',
                'start_date': '2020-03-29',
                'schedule': '30 23 * * *',
                'catchup': True,
                'engineers': [
                    {
                        'name': 'Engineer 1',
                        'slack': 'engineer_1'
                    }
                ],
                'tasks': [
                    {
                        'id': 'stage_1',
                        'operator': 'airflow.operators.bash_operator.BashOperator',
                        'params': {
                            'bash_command': 'echo "hello world"'
                        }
                    }
                ]
            },
            {
                'name': 'pipeline_2',
                'start_date': '2020-03-30',
                'schedule': None,
                'engineers': [
                    {
                        'name': 'Engineer 2',
                        'slack': 'engineer_2'
                    }
                ],
                'tasks': [
                    {
                        'id': 'stage_1',
                        'operator': 'airflow.operators.bash_operator.BashOperator',
                        'params': {
                            'bash_command': 'echo "hello world"'
                        }
                    }
                ]
            }
        ]

        self.hocon_configs = [
            {
                'name': 'pipeline_3',
                'start_date': '2020-03-29',
                'schedule': '30 23 * * *',
                'catchup': True,
                'engineers': [
                    {
                        'name': 'Engineer 3',
                        'slack': 'engineer_3'
                    }
                ],
                'tasks': [
                    {
                        'id': 'stage_1',
                        'operator': 'airflow.operators.bash_operator.BashOperator',
                        'params': {
                            'bash_command': 'echo "hello world"'
                        }
                    }
                ]
            },
            {
                'name': 'pipeline_4',
                'start_date': '2020-03-30',
                'schedule': None,
                'engineers': [
                    {
                        'name': 'Engineer 4',
                        'slack': 'engineer_4'
                    }
                ],
                'tasks': [
                    {
                        'id': 'stage_1',
                        'operator': 'airflow.operators.bash_operator.BashOperator',
                        'params': {
                            'bash_command': 'echo "hello world"'
                        }
                    }
                ]
            }
        ]

        self.invalid_config = [
            {
                'name': 'invalid_config',
                'start_date': '2020-03-29',
                'schedule': '30 23 * * *',
                'catchup': True,
                'engineers': [
                    {
                        'name': 'Engineer 3',
                        'slack': 'engineer_3'
                    }
                ]
            }
        ]

        prepare_clean_tmp_dir(folder_name='config_cache')
        self.configs_path = os.path.join(tmp_dir_path(), 'config_cache')
        ensure_dir_exists(self.configs_path)

        for config in self.yaml_configs:
            config_path = os.path.join(self.configs_path, '{}.yml'.format(config['name']))
            dict_to_yml(config, config_path)

        for config in self.hocon_configs:
            config_path = os.path.join(self.configs_path, '{}.hocon'.format(config['name']))
            dict_to_hocon(config, config_path)

        for config in self.invalid_config:
            config_path = os.path.join(self.configs_path, '{}.yml'.format(config['name']))
            dict_to_yml(config, config_path)

        self.config_cache = ConfigCache(self.configs_path)

    def test_cache_directory_path(self):
        self.assertEqual(self.config_cache.cache_directory_path, '{}/.cache'.format(self.configs_path))

    def test_bundle_path(self):
        self.assertEqual(self.config_cache.bundle_path, '{}/.cache/configs_bundle.json'.format(self.configs_path))

    def test_config_files_glob(self):
        self.assertEqual(len(self.config_cache.config_files_glob), 5)
        self.assertCountEqual([file.split('/')[-1] for file in self.config_cache.config_files_glob],
                              ['pipeline_1.yml', 'pipeline_2.yml', 'pipeline_3.hocon', 'pipeline_4.hocon', 'invalid_config.yml'])

    def test_refresh_cache(self):
        self.config_cache.refresh_cache()
        self.assertTrue(os.path.exists(self.config_cache.bundle_path))

        with open(self.config_cache.bundle_path) as json_file:
            configs_read_from_bundle = json.load(json_file)

        self.assertCountEqual(self.yaml_configs + self.hocon_configs, configs_read_from_bundle)

    def test_read_cache(self):
        with self.assertRaises(FileNotFoundError):
            self.config_cache.read_cache()

        self.config_cache.refresh_cache()
        configs = self.config_cache.read_cache(environment='prod')
        self.assertCountEqual(
            [Config(config, environment='prod') for config in self.yaml_configs] +
            [Config(config, environment='prod') for config in self.hocon_configs],
            configs)


class TestDagLoader(TestCase):
    def setUp(self):
        self.configs = [
            {
                'name': 'pipeline_a',
                'start_date': '2020-03-29',
                'schedule': '30 23 * * *',
                'catchup': True,
                'engineers': [
                    {
                        'name': 'Engineer 1',
                        'slack': 'engineer_1'
                    }
                ],
                'default_dag_args': {
                    'foo': 'bar'
                },
                'tasks': [
                    {
                        'id': 'stage_1',
                        'operator': 'airflow.operators.bash_operator.BashOperator',
                        'params': {
                            'bash_command': 'echo "hello world"'
                        }
                    },
                    {
                        'id': 'stage_2',
                        'operator': 'airflow.operators.bash_operator.BashOperator',
                        'params': {
                            'bash_command': 'echo "hello universe"'
                        },
                        'upstream': ['stage_1']
                    },
                    {
                        'id': 'stage_2b',
                        'operator': 'airflow.operators.bash_operator.BashOperator',
                        'params': {
                            'bash_command': 'echo "hello universe in parallel"'
                        },
                        'upstream': ['stage_1']
                    }
                ]
            },
            {
                'name': 'pipeline_b',
                'start_date': '2020-03-30',
                'engineers': [
                    {
                        'name': 'Engineer 2',
                        'slack': 'engineer_2'
                    }
                ],
                'dependencies': [
                    {
                        'id': 'sensor_1',
                        'operator': 'airflow.sensors.time_sensor.TimeSensor',
                        'params': {
                            'target_time': '2020-03-29 00:00:00'
                        }
                    },
                    {
                        'id': 'sensor_2',
                        'operator': 'airflow.sensors.time_sensor.TimeSensor',
                        'params': {
                            'target_time': '2020-03-29 00:00:00'
                        }
                    },
                ],
                'tasks': [
                    {
                        'id': 'stage_1b',
                        'operator': 'airflow.operators.bash_operator.BashOperator',
                        'params': {
                            'bash_command': 'echo "hello world"'
                        }
                    },
                    {
                        'id': 'stage_2c',
                        'operator': 'airflow.operators.bash_operator.BashOperator',
                        'params': {
                            'bash_command': 'echo "hello world"'
                        }
                    }
                ]
            },
            {
                'name': 'pipeline_c',
                'start_date': '2020-03-30',
                'engineers': [
                    {
                        'name': 'Engineer 3',
                        'slack': 'engineer_3'
                    }
                ],
                'pre_hooks': [
                    {
                        'id': 'pre_hook_1',
                        'operator': 'airflow.sensors.time_sensor.TimeSensor',
                        'params': {
                            'target_time': '2020-03-29 00:00:00'
                        }
                    },
                    {
                        'id': 'pre_hook_2',
                        'operator': 'airflow.sensors.time_sensor.TimeSensor',
                        'params': {
                            'target_time': '2020-03-29 00:00:00'
                        },
                        'upstream': ['pre_hook_1']
                    }
                ],
                'dependencies': [
                    {
                        'id': 'sensor_3',
                        'operator': 'airflow.sensors.time_sensor.TimeSensor',
                        'params': {
                            'target_time': '2020-03-29 00:00:00'
                        }
                    }
                ],
                'tasks': [
                    {
                        'id': 'stage_3',
                        'operator': 'airflow.operators.bash_operator.BashOperator',
                        'params': {
                            'bash_command': 'echo "hello world"'
                        }
                    },
                    {
                        'id': 'stage_4',
                        'operator': 'airflow.operators.bash_operator.BashOperator',
                        'params': {
                            'bash_command': 'echo "hello world"'
                        }
                    }
                ]
            },
            {
                'name': 'pipe_with_sub_params',
                'start_date': '2020-03-30',
                'engineers': [
                    {
                        'name': 'Engineer 3',
                        'slack': 'engineer_3'
                    }
                ],
                'params': {
                    'default': {
                        'param1': 'one'
                    }
                },
                'tasks': [
                    {
                        'id': 'task_1',
                        'operator': 'airflow.operators.bash_operator.BashOperator',
                        'params': {
                            'bash_command': 'echo "{{ params.param1 }} - {{ params.param2 }}"',
                            'params': {
                                'param2': 'two'
                            }
                        }
                    }
                ]
            }
        ]

        prepare_clean_tmp_dir(folder_name='config_cache')
        self.configs_path = os.path.join(tmp_dir_path(), 'config_cache')
        ensure_dir_exists(self.configs_path)

        for config in self.configs:
            config_path = os.path.join(self.configs_path, '{}.yml'.format(config['name']))
            dict_to_yml(config, config_path)

        configs_cache = ConfigCache(configs_path=self.configs_path)
        configs_cache.refresh_cache()

        self.dag_loader = DagLoader(
            project="test_project",
            configs_path=self.configs_path,
            event_handlers=[]
        )

    def test_load(self):
        dags = self.dag_loader.load(global_symbol_table=globals())
        for variable_name, dag in dags:
            self.assertEqual(globals()[variable_name], dag)
            # Verify the DAG is serializable
            SerializedDAG.to_dict(dag)

    def test_dag_id_prefix(self):
        self.dag_loader.dag_id_prefix = 'prefix_'
        dags = self.dag_loader.load(global_symbol_table=globals())

        self.assertEqual(len([dag for var_name, dag in dags if dag.dag_id == 'pipeline_a_v1']), 0)
        self.assertEqual(len([dag for var_name, dag in dags if dag.dag_id == 'prefix_pipeline_a_v1']), 1)
        self.assertEqual(len([dag for var_name, dag in dags if dag.dag_id == 'pipeline_b_v1']), 0)
        self.assertEqual(len([dag for var_name, dag in dags if dag.dag_id == 'prefix_pipeline_b_v1']), 1)

        self.assertEqual(len([dag for var_name, dag in dags if dag.dag_id == 'pipeline_c_v1']), 0)
        self.assertEqual(len([dag for var_name, dag in dags if dag.dag_id == 'prefix_pipeline_c_v1']), 1)

        self.assertEqual(len([dag for var_name, dag in dags if dag.dag_id == 'pipe_with_sub_params_v1']), 0)
        self.assertEqual(len([dag for var_name, dag in dags if dag.dag_id == 'prefix_pipe_with_sub_params_v1']), 1)

    def test_load_dag_dependencies(self):
        dags = self.dag_loader.load(global_symbol_table=globals())

        # Pipeline A
        # ==========
        dag = next(dag for var_name, dag in dags if dag.dag_id == 'pipeline_a_v1')

        task_ids = list(map(lambda task: task.task_id, dag.tasks))
        self.assertCountEqual(task_ids, ['start', 'stage_1', 'stage_2', 'stage_2b', 'success'])

        start = dag.get_task('start')
        stage_1 = dag.get_task('stage_1')
        stage_2 = dag.get_task('stage_2')
        stage_2b = dag.get_task('stage_2b')
        success = dag.get_task('success')

        self.assertCountEqual(list(map(lambda task: task.task_id, start.upstream_list)), [])
        self.assertCountEqual(list(map(lambda task: task.task_id, stage_1.upstream_list)), ['start'])
        self.assertCountEqual(list(map(lambda task: task.task_id, stage_2.upstream_list)), ['stage_1'])
        self.assertCountEqual(list(map(lambda task: task.task_id, stage_2b.upstream_list)), ['stage_1'])
        self.assertCountEqual(list(map(lambda task: task.task_id, success.upstream_list)), ['stage_2', 'stage_2b'])

        # Pipeline B
        # ==========
        dag = next(dag for var_name, dag in dags if dag.dag_id == 'pipeline_b_v1')

        task_ids = list(map(lambda task: task.task_id, dag.tasks))
        self.assertCountEqual(task_ids, ['start', 'sensor_1', 'sensor_2', 'stage_1b', 'stage_2c', 'success'])

        start = dag.get_task('start')
        sensor_1 = dag.get_task('sensor_1')
        sensor_2 = dag.get_task('sensor_2')
        stage_1b = dag.get_task('stage_1b')
        stage_2c = dag.get_task('stage_2c')
        success = dag.get_task('success')

        self.assertCountEqual(list(map(lambda task: task.task_id, start.upstream_list)), [])
        self.assertCountEqual(list(map(lambda task: task.task_id, sensor_1.upstream_list)), ['start'])
        self.assertCountEqual(list(map(lambda task: task.task_id, sensor_2.upstream_list)), ['start'])
        self.assertCountEqual(list(map(lambda task: task.task_id, stage_1b.upstream_list)), ['sensor_1', 'sensor_2'])
        self.assertCountEqual(list(map(lambda task: task.task_id, stage_2c.upstream_list)), ['sensor_1', 'sensor_2'])
        self.assertCountEqual(list(map(lambda task: task.task_id, success.upstream_list)), ['stage_1b', 'stage_2c'])

        # Pipeline C
        # ==========
        dag = next(dag for var_name, dag in dags if dag.dag_id == 'pipeline_c_v1')

        task_ids = list(map(lambda task: task.task_id, dag.tasks))
        self.assertCountEqual(task_ids,
                              ['pre_hook_1', 'pre_hook_2', 'start', 'sensor_3', 'stage_3', 'stage_4', 'success'])

        pre_hook_1 = dag.get_task('pre_hook_1')
        pre_hook_2 = dag.get_task('pre_hook_2')
        start = dag.get_task('start')
        sensor_3 = dag.get_task('sensor_3')
        stage_3 = dag.get_task('stage_3')
        stage_4 = dag.get_task('stage_4')
        success = dag.get_task('success')

        self.assertCountEqual(list(map(lambda task: task.task_id, pre_hook_1.upstream_list)), [])
        self.assertCountEqual(list(map(lambda task: task.task_id, pre_hook_2.upstream_list)), ['pre_hook_1'])
        self.assertCountEqual(list(map(lambda task: task.task_id, start.upstream_list)), ['pre_hook_2'])
        self.assertCountEqual(list(map(lambda task: task.task_id, sensor_3.upstream_list)), ['start'])
        self.assertCountEqual(list(map(lambda task: task.task_id, stage_3.upstream_list)), ['sensor_3'])
        self.assertCountEqual(list(map(lambda task: task.task_id, stage_4.upstream_list)), ['sensor_3'])
        self.assertCountEqual(list(map(lambda task: task.task_id, success.upstream_list)), ['stage_3', 'stage_4'])


    def test_load_dag_with_default_dag_params(self):
        dags = self.dag_loader.load(global_symbol_table=globals())

        dag = next(dag for var_name, dag in dags if dag.dag_id == 'pipeline_a_v1')
        self.assertEqual(dag.default_args['foo'], 'bar')


    def test_load_dag_with_sub_params(self):
        # pipe_with_sub_params_v1
        dags = self.dag_loader.load(global_symbol_table=globals())

        dag = next(dag for var_name, dag in dags if dag.dag_id == 'pipe_with_sub_params_v1')
        
        exec_date = pendulum.now(tz='UTC')
        dag.create_dagrun(state=DagRunState.RUNNING,
                                execution_date=exec_date,
                                start_date=pendulum.datetime(2022, 3, 1, tz='UTC'),
                                run_type=DagRunType.MANUAL)

        task_1 = dag.get_task('task_1')
        task_instance = TaskInstance(task=task_1, execution_date=exec_date)
        task_instance_context = task_instance.get_template_context()

        task_1.render_template_fields(task_instance_context)
        self.assertEqual(task_1.bash_command, 'echo "one - two"')


    def test_build_task(self):
        task = DagLoader.build_task(task_id='some_task__id',
                                    dag=DAG(dag_id='some_dag', default_args={'start_date': datetime.datetime.now()}),
                                    operator='airflow.operators.bash_operator.BashOperator',
                                    bash_command='echo "hello world"')

        self.assertIsInstance(task, BashOperator)
        self.assertEqual(task.task_id, 'some_task__id')
        self.assertEqual(task.bash_command, 'echo "hello world"')

    def test_build_task_with_expected_class(self):
        task = DagLoader.build_task(task_id='some_task__id',
                                    dag=DAG(dag_id='some_dag', default_args={'start_date': datetime.datetime.now()}),
                                    operator='airflow.operators.bash_operator.BashOperator',
                                    expected_class=BashOperator,
                                    bash_command='echo "hello world"')
        self.assertIsInstance(task, BashOperator)

        task = DagLoader.build_task(task_id='some_task__id',
                                    dag=DAG(dag_id='some_dag', default_args={'start_date': datetime.datetime.now()}),
                                    operator='airflow.operators.bash_operator.BashOperator',
                                    expected_class=BaseOperator,
                                    bash_command='echo "hello world"')
        self.assertIsInstance(task, BashOperator)

        with self.assertRaises(Exception):
            task = DagLoader.build_task(task_id='some_task__id',
                                        dag=DAG(dag_id='some_dag',
                                                default_args={'start_date': datetime.datetime.now()}),
                                        operator='airflow.operators.bash_operator.BashOperator',
                                        expected_class=PythonOperator,
                                        bash_command='echo "hello world"')


class TestTasksSubDag(TestCase):

    def setUp(self):
        self.task_configs = [
            {
                'id': 'stage_1',
                'operator': 'airflow.operators.bash_operator.BashOperator',
                'params': {
                    'bash_command': 'echo "hello world"'
                }
            },
            {
                'id': 'stage_2',
                'operator': 'airflow.operators.bash_operator.BashOperator',
                'params': {
                    'bash_command': 'echo "hello universe"'
                },
                'upstream': ['stage_1']
            },
            {
                'id': 'stage_2b',
                'operator': 'airflow.operators.bash_operator.BashOperator',
                'params': {
                    'bash_command': 'echo "hello universe in parallel"'
                },
                'upstream': ['stage_1']
            }
        ]

        self.dag = DAG(dag_id='some_dag', default_args={'start_date': datetime.datetime.now()})

        self.sub_dag = TasksSubDag(
            task_configs=self.task_configs,
            dag=self.dag,
            job_name='some_job_name',
            params={}
        )

    def test_leaves(self):
        leaf_ids = [leaf.task_id for leaf in self.sub_dag.leaves]

        self.assertCountEqual(leaf_ids, ['stage_2', 'stage_2b'])

    def test_roots(self):
        root_ids = [root.task_id for root in self.sub_dag.roots]

        self.assertCountEqual(root_ids, ['stage_1'])


class TestDependencySensors(TestCase):

    def setUp(self):
        self.sensor_configs = [
            {
                'id': 'sensor_1',
                'operator': 'airflow.sensors.time_sensor.TimeSensor',
                'params': {
                    'target_time': datetime.time(0, 0, 0)
                }
            },
            {
                'id': 'sensor_2',
                'operator': 'airflow.sensors.time_sensor.TimeSensor',
                'params': {
                    'target_time': datetime.time(0, 0, 0)
                }
            },
            {
                'id': 'sensor_3',
                'operator': 'airflow.sensors.time_sensor.TimeSensor',
                'params': {
                    'target_time': datetime.time(0, 0, 0)
                }
            }
        ]

        self.dependent_sensor_configs = [
            {
                'id': 'sensor_4',
                'operator': 'airflow.sensors.time_sensor.TimeSensor',
                'params': {
                    'target_time': datetime.time(0, 0, 0)
                }
            },
            {
                'id': 'sensor_5',
                'operator': 'airflow.sensors.time_sensor.TimeSensor',
                'params': {
                    'target_time': datetime.time(0, 0, 0)
                },
                'upstream': ['sensor_4']
            },
            {
                'id': 'sensor_6',
                'operator': 'airflow.sensors.time_sensor.TimeSensor',
                'params': {
                    'target_time': datetime.time(0, 0, 0)
                },
                'upstream': ['sensor_4']
            }
        ]

        self.dag = DAG(dag_id='some_dag', default_args={'start_date': datetime.datetime.now()})

        self.dependency_sensors = DependencySensors(
            task_configs=self.sensor_configs,
            dag=self.dag,
            params={}
        )

        self.dependent_dependency_sensors = DependencySensors(
            task_configs=self.dependent_sensor_configs,
            dag=self.dag,
            params={}
        )

    def test_leaves(self):
        task_ids = [task.task_id for task in self.dependency_sensors.leaves]
        self.assertCountEqual(task_ids, ['sensor_1', 'sensor_2', 'sensor_3'])

    def test_roots(self):
        task_ids = [task.task_id for task in self.dependency_sensors.roots]
        self.assertCountEqual(task_ids, ['sensor_1', 'sensor_2', 'sensor_3'])

    def test_dependent_leaves(self):
        task_ids = [task.task_id for task in self.dependent_dependency_sensors.leaves]
        self.assertCountEqual(task_ids, ['sensor_5', 'sensor_6'])

    def test_dependent_roots(self):
        task_ids = [task.task_id for task in self.dependent_dependency_sensors.roots]
        self.assertCountEqual(task_ids, ['sensor_4'])

    def test_timeout_set_to_zero(self):
        for sensor in self.dependency_sensors.tasks.values():
            self.assertEqual(sensor.timeout, 0)





import click
import logging
import os

from awehflow_cli.config_scripts.generate_configs import generate_config
from awehflow_cli.config_scripts.validate_configs import validate_config

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(message)s')


@click.group(help="CLI tool for awehflow")
def aweh():
    """ main command """
    pass

@aweh.command()
@click.option("-sql_file", type=str, required=True, help="The sql file")
@click.pass_context
def gencon(ctx, sql_file):
    """Generates index.yml config from a provided .sql file"""
    from pathlib import Path
    current_dir = Path(os.getcwd())
    sql_files_list = list(current_dir.glob('*.sql'))
    if len(sql_files_list)==0:
        click.echo("Please navigate to directory that contains the sql file")
        ctx.exit(1)
    else:
        usecase = click.prompt("What is the usecase? e.g. client_transactions", type=str)
        start_date = click.prompt("What is the start date? e.g. 2025-01-01", type=str)

        schedule = click.prompt(
            "Choose a schedule for the DAG e.g. 0 * * * * (hourly), 0 3 * * * (daily @ 03:00), 0 3 1 * * (monthly on the 1st @ 03:00)",
            type=str
        )

        source_project = click.prompt("Enter the source project", type=str)
        source_dataset = click.prompt("Enter the source dataset", type=str)

        target_project = click.prompt("Enter the target project", type=str)
        target_dataset = click.prompt("Enter the target dataset", type=str)

        partition = click.prompt("Would you like to partition your table?", type=click.Choice(['Y', 'n']), default='Y', show_default=True)

        if partition =="Y":
            partition = {
                "partition_field": click.prompt("What field would you like to partition on?", type=str),
                "partition_type": click.prompt("How would you like to partition?", type=click.Choice(['HOUR', 'DAY', 'MONTH', 'YEAR']))
            }
        else:
            partition = None

        cluster = click.prompt("Would you like to add clustering to your table?", type=click.Choice(['Y', 'n']), default='Y', show_default=True)

        if cluster =="Y":
            cluster_fields = click.prompt(
                "What fields would you like to cluster on? e.g. id, date. NB: BigQuery allows maximum 4 clustering fields.",
                type=str
            )
        else:
            cluster_fields = None

        generate_config(sql_file, usecase, start_date, schedule, source_project, source_dataset, target_project, target_dataset, partition, cluster_fields)


@aweh.command()
@click.option("-path_to_configs_dir", type=str, required=True, help="Specify the path to the configs directory e.g. dags/prepared/configs")
def validate(path_to_configs_dir):
    """Validates all the index.yml configs against their respective .sql files"""
    validate_config(path_to_configs_dir)


if __name__ == '__main__':
    aweh()


from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook
from airflow.providers.postgres.hooks.postgres import PostgresHook
from airflow.hooks.base import BaseHook
from airflow.models.baseoperator import BaseOperator
from airflow.utils.decorators import apply_defaults
from airflow.providers.ssh.hooks.ssh import SSHHook
from pathlib import Path
from datetime import datetime as dt
from random import randint
import time
from typing import List
import fnmatch
import os


class BigQueryCSVExtractAndComposeOperator(BaseOperator):
    """
    Uses Google Cloud Storage compose to extract a BigQuery table to a single CSV single file
    """

    template_fields = (
        'source_project',
        'source_dataset',
        'source_table',
        'destination_object_name',
        'temp_bucket_name', 
        'temp_dataset_name', )

    @apply_defaults
    def __init__(self,
                 source_project: str,
                 source_dataset: str,
                 source_table: str,
                 destination_object_name: str,
                 temp_bucket_name: str,
                 temp_dataset_name: str,
                 print_header: bool = True,
                 field_delimiter: str = ',',
                 bigquery_conn_id: str = 'bigquery_default',
                 gcp_conn_id: str = 'gcp_conn_id',
                 location=None,
                 *args, **kwargs) -> None:

        """
        Create an instance of the BigQueryCSVExtractAndComposeOperator
        :param source_dataset_table: The source BigQuery table to extract, including the data set and table name ie. dev_temp.some_table_to_extract
        :param destination_object_name: The destination filename as a bucket URI ie gs://dev_extract_data/some_sub_folder/my_extract.csv
        :param temp_bucket_name: The temporary bucket area where files can be extracted and composed
        :param temp_dataset_name: A temp dataset where temporary tables can be created and saved, while processing (can also pass in project as <project>.<dataset>)
        :param print_header: Whether or not the extracted data should contain a header or not
        :param field_delimiter: What kind of field delimiter to use. Default: ,
        :param bigquery_conn_id: The name of the google_cloud_platform connection to use for queries.  The user should have R/W permission to the temp dataset
        :param gcp_conn_id: The name of the google_cloud_platform connection to use for GCS jobs.  The connection should have R/W permission to the destination bucket and temp bucket
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)

        self.source_project = source_project
        self.source_dataset = source_dataset
        self.source_table = source_table
        self.destination_object_name = destination_object_name
        self.temp_bucket_name = temp_bucket_name.strip('/')
        self.temp_dataset_name = temp_dataset_name
        self.print_header = print_header
        self.field_delimiter = field_delimiter
        self.bigquery_conn_id = bigquery_conn_id
        self.gcp_conn_id = gcp_conn_id
        self.location = location

        # place holders for execute method
        self.gcs_hook = None
        self.bq_hook = None

    def execute(self, context):
        """
        Execute the BigQueryCSVExtractAndComposeOperator to extract a biq query table to a bucket and then compose the parts into a single CSV file
        :param context: The airflow context to execute on
        :return:
        """

        if self.gcs_hook is None:
            self.log.info('Creating GCSHook')
            self.gcs_hook = GCSHook(gcp_conn_id=self.gcp_conn_id)
            credentials = self.gcs_hook.get_credentials()
            credential_email = credentials.service_account_email
            self.log.info(f"Created GCSHook with account: {credential_email}")

        if self.bq_hook is None:
            self.log.info('Creating BigQueryHook')
            self.bq_hook = BigQueryHook(gcp_conn_id=self.bigquery_conn_id, location=self.location)
            credentials = self.bq_hook.get_credentials()
            credential_email = credentials.service_account_email
            self.log.info(f"Created BigQueryHook with account: {credential_email}")

        compose_bucket_name, compose_filename = self.get_bucket_and_filename(object_name=self.destination_object_name)
        self.log.info('Calculated final compose_bucket_name [{}] and compose_filename [{}]'.format(compose_bucket_name, compose_filename))

        working_bucket_name, junk_data = self.get_bucket_and_filename(object_name=self.temp_bucket_name + '/junk.dat')
        self.log.info('Calculated working_bucket_name: {}'.format(working_bucket_name))

        tmp_file_uniq = 'data_extract_tmp_' + dt.now().strftime('%Y%m%H%M%S') + '_' + str(randint(1, 100))
        tmp_parts_name = tmp_file_uniq + '_*.csv'
        tmp_compose_file = 'composed_data_' + tmp_file_uniq + '.csv'

        try:
            if self.print_header:
                self.log.info('Header requested so generating header tbl and file')

                tmp_header_filename = self.temp_bucket_name + '/_header_' + tmp_file_uniq + '_*.csv'
                tmp_header_table_name = '{}.{}_header_data'.format(self.temp_dataset_name, tmp_file_uniq)
                tmp_split = tmp_header_table_name.replace(":", ".").split(".")
                if len(tmp_split) == 3:
                    project_id, dataset_id, table_id = tmp_split
                    self.log.info(f'Getting project_id from temp_dataset_name argument: {project_id}.{dataset_id}.{table_id}')

                else:
                    dataset_id, table_id = tmp_split
                    project_id = self.bq_hook.project_id
                    self.log.info(f'Getting project_id from bq_hook {project_id}.{dataset_id}.{table_id}')


                self.log.info('Defined tmp_header_filename [{}] and tmp_header_table_name [{}]'.format(tmp_header_filename, tmp_header_table_name))
                self.log.info('Creating empty header table [{}]'.format(tmp_header_table_name))

                header_table_created = False

                try:
                    header_job_id = self.bq_hook.insert_job(
                        configuration = {
                            "query": {
                                "query": f'SELECT * FROM `{self.source_project}.{self.source_dataset}.{self.source_table}` WHERE 1=0 LIMIT 1',
                                "useLegacySql": False,
                                "writeDisposition": "WRITE_TRUNCATE",
                                "destinationTable": {
                                    "projectId": project_id,
                                    "datasetId": dataset_id,
                                    "tableId": table_id
                                }
                            }
                        },
                        location=self.location
                    )

                    header_job = self.bq_hook.get_job(job_id=header_job_id)
                    while not header_job.done():
                        self.log.info(f'Waiting for header query [{header_job_id}]...')
                        time.sleep(1)

                    header_table_created = True

                    self.log.info('Exporting header table [{}] to file [{}]'.format(tmp_header_table_name, tmp_header_filename))
                    
                    extract_job_id = self.bq_hook.insert_job(
                        configuration={
                            "extract": {
                                "sourceTable": {
                                    "projectId": project_id,
                                    "datasetId": dataset_id,
                                    "tableId": table_id
                                },
                                "destinationUris": [f'gs://{working_bucket_name}/{tmp_header_filename}'],
                                "printHeader": True,
                                "fieldDelimiter": self.field_delimiter,
                                "destinationFormat": "CSV" 
                            }
                        }
                    )

                    extract_job = self.bq_hook.get_job(job_id=extract_job_id)
                    while not extract_job.done():
                        self.log.info(f'Waiting for extract job [{extract_job_id}]...')
                        time.sleep(1)

                    self.log.info('Completed export of header table data')
                finally:
                    if header_table_created:
                        try:
                            self.bq_hook.delete_table(table_id=f'{self.source_project}.{tmp_header_table_name}', not_found_ok=True)
                        except ValueError:
                            self.bq_hook.delete_table(table_id=tmp_header_table_name, not_found_ok=True) 

            self.log.info("Extracting table data to tmp_parts_name [{}]".format(tmp_parts_name))
            final_extract_job_id = self.bq_hook.insert_job(
                configuration = {
                    "extract": {
                        "sourceTable": {
                                "projectId": self.source_project,
                                "datasetId": self.source_dataset,
                                "tableId": self.source_table,
                            },
                        "destinationUris": [f'gs://{working_bucket_name}/{tmp_parts_name}'],
                        "printHeader": False,
                        "fieldDelimiter": self.field_delimiter,
                        "destinationFormat": "CSV" 
                    }
                }
            )

            final_extract_job = self.bq_hook.get_job(job_id=final_extract_job_id)
            while not final_extract_job.done():
                self.log.info(f'Waiting for final extract job [{final_extract_job_id}]...')
                time.sleep(1)

            self.log.info('Completed extraction of primary table')
            self.log.info('Listing file parts matching [{}] to compose'.format(tmp_file_uniq))

            source_list = self.list_extract_files(working_bucket_name, tmp_file_uniq)
            self.log.info('Composing tmp_compose_file from source_list')
            self.log.info('source_list: {}'.format(source_list))
            self.log.info('-->')
            self.log.info('tmp_compose_file: {}'.format(tmp_compose_file))

            self.gcs_hook.compose(working_bucket_name,
                         source_objects=source_list,
                         destination_object=tmp_compose_file
                )

            self.log.info('Copying tmp_compose_file [{}/{}] to final compose_filename [{}/{}]'.format(
                self.temp_bucket_name,
                tmp_compose_file,
                compose_bucket_name,
                compose_filename
            ))

            self.gcs_hook.copy(source_bucket=working_bucket_name, source_object=tmp_compose_file,
                               destination_bucket=compose_bucket_name,
                               destination_object=compose_filename)

            self.log.info('Completed header-less extract & compose')
        finally:
            self.log.info('Deleting temp files created in [{}] during export process'.format(working_bucket_name))
            source_list = self.list_extract_files(bucket_name=working_bucket_name, file_pattern=tmp_file_uniq)
            for src in source_list:
                self.log.info('Deleting file [{}]'.format(src))
                self.gcs_hook.delete(working_bucket_name, src) # pragma: no cover

    def get_bucket_and_filename(self, object_name):
        """
        Split the gs:// URI for the destination object to a bucket_name and file_name
        :param object_name: The cloud storage URI
        :return: a tuple containing two variables, the bucket_name and the file_name component
        """
        bucket_path = Path(object_name.replace('gs:/', ''))
        bucket_name = str(bucket_path.parent).split('/')[1]
        dest_filename = object_name.replace('gs://' + bucket_name + '/', '')
        return bucket_name, dest_filename

    def list_extract_files(self, bucket_name, file_pattern):
        source_list = []

        file_list = self.gcs_hook.list(bucket_name)
        for filename in file_list:

            if file_pattern in str(filename):
                self.log.info('Found tmp file [{}], adding to source list'.format(filename))

                if 'header' in str(filename):
                    source_list.insert(0,filename)
                else:
                    source_list.append(filename)

        return source_list

class RecursiveComposeGCSFilesOperator(BaseOperator):
    """
    Recursively composes files from a Google Cloud Storage bucket until there's only one file left.
    Allows to compose more than 32 files, which is the GCS limit

    :param gcp_conn_id: The connection ID to use when connecting to Google Cloud Platform.
    :type gcp_conn_id: str
    :param bucket_name: The name of the Google Cloud Storage bucket.
    :type bucket_name: str
    :param source_prefix: The prefix for the source objects in the bucket. F.e 'transactions_2024'
    :type source_prefix: str
    :param destination_blob: The name of the composed destination blob INCLUDING FILE TYPE. F.e 'transactions.csv'
    :type destination_blob: str
    """
    
    template_fields = ['bucket_name', 'source_prefix', 'destination_blob'] # Allows airflow variable replacement
    
    def __init__(
            self, 
            gcp_conn_id: str = None, 
            bucket_name: str = None,
            source_prefix: str = None,
            destination_blob: str = None,
            *args, **kwargs
        ):
        super(RecursiveComposeGCSFilesOperator, self).__init__(*args, **kwargs)
        self.gcp_conn_id = gcp_conn_id
        self.bucket_name = bucket_name
        self.source_prefix = source_prefix
        self.destination_blob = destination_blob

    def rename_blob(self, blob_name):
        # Copy the source blob to the destination
        self.gcs_hook.copy(source_bucket=self.bucket_name, source_object=blob_name, destination_object=self.destination_blob)

        # Delete the original blob
        self.gcs_hook.delete(bucket_name=self.bucket_name, object_name=blob_name)
        self.log.info(f'Blob {blob_name} renamed to {self.destination_blob}')


    def compose_files(self, files: List[str]):
        """
        Recursively composes files from the given list until there's only one file left.

        :param files: List of file names to compose.
        :type files: List[str]
        """
        self.log.info("Composing files: %s", files)

        files_to_compose = files[:32]  # Google Cloud Storage has a limit of 32 files per compose operation

        # {self.source_prefix} in destination_object ensures that the dest blob adheres to source_prefix for recursive compose
        # _{len(files)}' ensures unique name and doesn't try to overwrite another object
        # {self.destination_blob.rsplit('.')[-1]} maintains destination object type
        self.gcs_hook.compose(bucket_name=self.bucket_name, source_objects=files_to_compose, destination_object=f'{self.source_prefix}_{len(files)}.{self.destination_blob.rsplit(".")[-1]}')
        self.log.info("Composed files into: %s", f'{self.source_prefix}_{len(files)}.{self.destination_blob.rsplit(".")[-1]}')
        
        # Remove the files that were composed
        for file_name in files_to_compose:
            self.log.info("Deleting file: %s", file_name)
            self.gcs_hook.delete(bucket_name=self.bucket_name, object_name=file_name)

        # Recursively call compose_files with remaining files - includes previously composed blob(s)
        remaining_blobs = list(self.gcs_hook.get_conn().bucket(self.bucket_name).list_blobs(prefix=self.source_prefix))
        remaining_files = [blob.name for blob in remaining_blobs]

        if len(remaining_files) == 1:
            self.log.info("Only one file left. Compose operation completed.")
            return remaining_files
        
        return self.compose_files(remaining_files)

    def execute(self, context):
        self.gcs_hook = GCSHook(gcp_conn_id=self.gcp_conn_id)
        blobs = list(self.gcs_hook.get_conn().bucket(self.bucket_name).list_blobs(prefix=self.source_prefix))

        if len(blobs) == 1:
            self.log.info("1 file found. No need to compose.")
            self.log.info(blobs[0].name)
            self.rename_blob(blobs[0].name)
        elif len(blobs) == 0:
            self.log.info("0 files found. Nothing to compose.")
        else:
            files = [blob.name for blob in blobs]
            composed_blob = self.compose_files(files)
            self.log.info(composed_blob[0])
            self.rename_blob(composed_blob[0])

class SftpToGcsOperator(BaseOperator):
    template_fields = ['source_folder', 'file_string', 'gcs_bucket_name']
    """
    A custom Airflow operator that extracts csv files from an sftp server and drops it on a GCS bucket. 
    If the optional 'stateful' flag is set to true, this operator writes out metadata about the files extracted 
    from the sftp server and ignores any duplicates in subsequent runs. If this option is set to true, 
    the postgres_conn_id is required. The metadata table should be initialised using the following sql:

    CREATE TABLE public.sftp_metrics (
        host varchar NOT NULL,
        file_path varchar NOT NULL,
        destination_path varchar NOT NULL,
        processed_time timestamp NOT NULL,
        status varchar NOT NULL,

        CONSTRAINT pk_sftp_metrics PRIMARY KEY (host, file_path)
    );

    :param gcp_conn_id: The connection ID to use when connecting to Google Cloud Platform.
    :type gcp_conn_id: str
    :param sftp_conn_id: The connection ID to use when connecting to the SFTP server.
    :type sftp_conn_id: str
    :param source_folder: The source folder on sftp server where the files are located (None if in user home directory).
    :type source_folder: str
    :param file_string: The file string to look and match target file(s) with (use * for wildcard)
    :type file_string: str
    :param gcs_bucket_name: The name of the gcs bucket to move the file(s) to
    :type gcs_bucket_name: str
    :param dest_folder: The destination folder to write file(s) to in gcs bucket
    :type dest_folder: str
    :param move_object: Set to True to delete source file(s)
    :type move_object: bool
    :param fail_on_missing_file: Set to True to fail when no file(s) are matched on sftp server
    :type move_object: bool
    :param stateful: Set to True to write to metadata table, and ignore duplicate files
    :type stateful: bool
    :param postgres_conn_id: Airflow postgres connection that metadata is written out to
    :type postgres_conn_id: str
    """

    @apply_defaults
    def __init__(
        self,
        gcp_conn_id: str = None,
        sftp_conn_id: str = None,
        source_folder: str = '',
        file_string: str = None,
        gcs_bucket_name: str = None,
        dest_folder: str = '',
        move_object: bool = False,
        fail_on_missing_file: bool = False,
        stateful: bool = False,
        postgres_conn_id: str = '',
        metadata_table_name: str = 'sftp_metrics',
        *args, **kwargs
    ):
        super(SftpToGcsOperator, self).__init__(*args, **kwargs)
        self.gcp_conn_id = gcp_conn_id
        self.sftp_conn_id = sftp_conn_id
        self.source_folder = source_folder
        self.file_string = file_string
        self.gcs_bucket_name = gcs_bucket_name
        self.dest_folder = dest_folder
        self.move_object = move_object
        self.fail_on_missing_file = fail_on_missing_file
        self.stateful = stateful
        self.postgres_conn_id = postgres_conn_id
        self.metadata_table_name = metadata_table_name

    def _file_in_metastore(self, file):
        """
        Helper function that returns true if the file has already been processed
        """

        file_path = os.path.join(self.source_folder, file)
        postgres_hook = PostgresHook(postgres_conn_id=self.postgres_conn_id)

        sql = f"""
        SELECT 1 
        FROM {self.metadata_table_name} 
        WHERE file_path = %s AND host = %s
        """
        
        result = postgres_hook.get_first(sql, (file_path, self.sftp_host))
        return result is not None
        
    def _get_file(self, sftp):
        files_on_server = sftp.listdir(self.source_folder)
        matching_files = [file for file in files_on_server if fnmatch.fnmatch(file, self.file_string)]
        paths = []

        if len(matching_files) > 0:
            for file in matching_files:
                # If the operator is not stateful it will return all files that matches the file string
                if not self.stateful:
                    paths.append(os.path.join(self.source_folder, file))
                    self.log.info(f"Found matching file: {file}")
                else:
                    if not self._file_in_metastore(file):
                        paths.append(os.path.join(self.source_folder, file))
                        self.log.info(f"Found matching file: {file}")
                    else:
                        self.log.info(f"Skipping already processed matching file: {file}")
                        
        return paths

    def _update_metastore(self, sftp_host, file_path, output_object, status):
        """
        Inserts the metadata for a processed file to postgres
        """

        sql = f"""
            INSERT INTO {self.metadata_table_name} (host, file_path, destination_path, processed_time, status)
            VALUES (%s, %s, %s, CURRENT_TIMESTAMP, %s)
            ON CONFLICT (host, file_path) 
            DO UPDATE SET 
                destination_path = EXCLUDED.destination_path,
                processed_time = EXCLUDED.processed_time,
                status = EXCLUDED.status;
        """
        
        self.pg_hook.run(sql, parameters=(sftp_host, file_path, output_object, status))
        self.log.info(f"Metastore updated for: {file_path}")

    def execute(self, context):
        try:
            gcs_hook = GCSHook(gcp_conn_id=self.gcp_conn_id)
            bucket = gcs_hook.get_conn().bucket(self.gcs_bucket_name)
            ssh_hook = SSHHook(ssh_conn_id=self.sftp_conn_id)
            self.sftp_host = ssh_hook.remote_host
            self.pg_hook = PostgresHook(postgres_conn_id=self.postgres_conn_id)

            with ssh_hook.get_conn() as ssh_client:
                with ssh_client.open_sftp() as sftp_client:
                    report_object_paths = self._get_file(sftp_client)
                    if len(report_object_paths) < 1 and self.fail_on_missing_file:
                        raise ValueError("No matching files found")

                    for path in report_object_paths:
                        output_object = os.path.join(self.dest_folder or '', path.split("/")[-1])

                        with sftp_client.open(path, bufsize=32768) as remote_file:
                            blob = bucket.blob(f'{output_object}', chunk_size=262144)
                            blob.upload_from_file(remote_file)
                            if self.stateful:
                                self._update_metastore(sftp_host=self.sftp_host, file_path=path, output_object="gs://"+self.gcs_bucket_name+"/"+output_object, status="COPIED")
                        self.log.info(f'File: {output_object} uploaded successfully')
                        if self.move_object:
                            sftp_client.remove(path)
                            self.log.info(f'File {path} deleted successfully.')
                            if self.stateful:
                                self._update_metastore(sftp_host=self.sftp_host, file_path=path, output_object="gs://"+self.gcs_bucket_name+"/"+output_object, status="MOVED")

        except Exception as e:
            self.log.error(f"An error occurred during upload: {e}")
            raise
        finally:
            if 'sftp_client' in locals() and sftp_client:
                sftp_client.close()
            if 'ssh_client' in locals() and ssh_client:
                ssh_client.close()